(********************************************************)
(*   Olivier Boite (under supervision C. Dubois) - 2001 *)
(*   Catherine Dubois - translation to V8 - sept. 2012  *)
(********************************************************)

Require Import Language_spec. 
Require Import Table.
Require Import Typing_spec.


Lemma equiv_in_type_of :
 forall (e : expr) (t : type) (envx envx' : env_typ_var) (envl : env_typ_loc),
 equiv envx envx' -> type_of (envx', envl) e t -> type_of (envx, envl) e t.

intros e.
induction e
 as
  [c|
   i|
   l|
   i e Hrece|
   e1 Hrece1 e0 Hrece0|
   i e1 Hrece1 e0 Hrece0|
   e1 Hrece1 e0 Hrece0 e2 Hrece2|
   e1 Hrece1 e0 Hrece0|
   e Hrece|
   e Hrece|
   e1 Hrece1 e0 Hrece0|
   e1 Hrece1 e0 Hrece0]; intros t envx envx' envl Hequiv Htyp; 
 inversion Htyp.
apply type_of_const.

apply type_of_var.
apply apply_equiv with (t' := envx'); assumption.

apply type_of_loc; assumption.

apply type_of_fun.
apply Hrece with (envx' := add_table i t0 envx').
apply eqhead; assumption.
assumption.

apply type_of_app with (t := t0).
apply Hrece1 with (envx' := envx'); assumption.
apply Hrece0 with (envx' := envx'); assumption.

apply type_of_let with (t := t0).
apply Hrece1 with (envx' := envx'); assumption.
apply Hrece0 with (envx' := add_table i t0 envx').
apply eqhead; assumption.
assumption.

apply type_of_if.
apply Hrece1 with (envx' := envx'); assumption.
apply Hrece0 with (envx' := envx'); assumption.
apply Hrece2 with (envx' := envx'); assumption.

apply type_of_while with (t := t0).
apply Hrece1 with (envx' := envx'); assumption.
apply Hrece0 with (envx' := envx'); assumption.

apply type_of_ref.
apply Hrece with (envx' := envx'); assumption.

apply type_of_deref.
apply Hrece with (envx' := envx'); assumption.

apply type_of_sequ with (t' := t').
apply Hrece1 with (envx' := envx'); assumption.
apply Hrece0 with (envx' := envx'); assumption.

apply type_of_assign with (t := t0).
apply Hrece1 with (envx' := envx'); assumption.
apply Hrece0 with (envx' := envx'); assumption.
Qed.


(*-------------------------------------*)

Lemma env_useless :
 forall (e : expr) (x : identifier) (t t' : type) (envx : env_typ_var)
   (envl : env_typ_loc),
 ~ free_ident x e ->
 type_of (envx, envl) e t' -> type_of (add_table x t envx, envl) e t'.

intros e.
induction e
 as
  [c|
   i|
   l|
   i e Hrece|
   e1 Hrece1 e0 Hrece0|
   i e1 Hrece1 e0 Hrece0|
   e1 Hrece1 e0 Hrece0 e2 Hrece2|
   e1 Hrece1 e0 Hrece0|
   e Hrece|
   e Hrece|
   e1 Hrece1 e0 Hrece0|
   e1 Hrece1 e0 Hrece0]; intros x t t' envx envl Hnot_free Htyp;
 inversion Htyp.

(*const*)
apply type_of_const.

(*var*)
case (eq_A_dec x i).
intros x_is_i.
absurd (free_ident x (Var i)).
assumption.

rewrite x_is_i.
apply free_var.

intro x_isnt_i.
apply type_of_var.
apply apply_add_non; assumption.

(*loc*)
apply type_of_loc; assumption.

(*fun*)
apply type_of_fun.
case (eq_A_dec x i).
intros x_is_i; rewrite x_is_i.
apply equiv_in_type_of with (envx' := add_table i t0 envx).
apply equseless; apply eqrefl.
assumption.

intro x_isnt_i.
apply equiv_in_type_of with (add_table x t (add_table i t0 envx)).
apply eqswap; auto; apply eqrefl.
apply Hrece.
intro Hfree.
absurd (free_ident x (Fun i e)).
assumption.

apply free_fun; assumption.

assumption.

(*apply*)
apply type_of_app with (t := t0).
apply Hrece1.
intro Hfree.
absurd (free_ident x (Apply e1 e0)).
assumption.

apply free_apply1; assumption.

assumption.

apply Hrece0.
intro Hfree.
absurd (free_ident x (Apply e1 e0)).
assumption.

apply free_apply2; assumption.

assumption.

(*let*)
apply type_of_let with (t := t0).
apply Hrece1.
intro Hfree.
absurd (free_ident x (Let_in i e1 e0)).
assumption.
apply free_let1; assumption.
assumption.

case (eq_A_dec x i).
intros x_is_i; rewrite x_is_i.
apply equiv_in_type_of with (envx' := add_table i t0 envx).
apply equseless; apply eqrefl. 
assumption.

intro x_isnt_i.
apply equiv_in_type_of with (add_table x t (add_table i t0 envx)).
apply eqswap; auto; apply eqrefl.

apply Hrece0.
intro Hfree.
absurd (free_ident x (Let_in i e1 e0)).
assumption.

apply free_let2; assumption.

assumption.

(*if*)
apply type_of_if.
apply Hrece1.
intro Hfree.
absurd (free_ident x (If e1 e0 e2)).
assumption.
apply free_if1; assumption.
assumption.

apply Hrece0.
intro Hfree.
absurd (free_ident x (If e1 e0 e2)).
assumption.
apply free_if2; assumption.
assumption.

apply Hrece2.
intro Hfree.
absurd (free_ident x (If e1 e0 e2)).
assumption.
apply free_if3; assumption.
assumption.

(*while*)
apply type_of_while with (t := t0).
apply Hrece1.
intro Hfree.
absurd (free_ident x (While e1 e0)).
assumption.
apply free_while1; assumption.
assumption.

apply Hrece0.
intro Hfree.
absurd (free_ident x (While e1 e0)).
assumption.
apply free_while2; assumption.
assumption.

(*ref*)
apply type_of_ref.
apply Hrece.
intro Hfree.
absurd (free_ident x (Ref e)).
assumption.
apply free_ref; assumption.
assumption.

(*deref*)
apply type_of_deref.
apply Hrece.
intro Hfree.
absurd (free_ident x (Deref e)).
assumption.
apply free_deref; assumption.
assumption.

(*sequ*)
apply type_of_sequ with (t' := t'0).
apply Hrece1.
intro Hfree.
absurd (free_ident x (Sequ e1 e0)).
assumption.
apply free_sequ1; assumption.
assumption.

apply Hrece0.
intro Hfree.
absurd (free_ident x (Sequ e1 e0)).
assumption.
apply free_sequ2; assumption.
assumption.

(*assign*)
apply type_of_assign with (t := t0).
apply Hrece1.
intro Hfree.
absurd (free_ident x (Assign e1 e0)).
assumption.
apply free_assign1; assumption.
assumption.

apply Hrece0.
intro Hfree.
absurd (free_ident x (Assign e1 e0)).
assumption.
apply free_assign2; assumption.
assumption.

Qed.



(**************************************************)
Lemma env_redundant :
 forall (e : expr) (x : identifier) (t t' : type) (envx : env_typ_var)
   (envl : env_typ_loc),
 ~ free_ident x e ->
 type_of (add_table x t envx, envl) e t' -> type_of (envx, envl) e t'.


intros e.
induction e
 as
  [c|
   i|
   l|
   i e Hrece|
   e1 Hrece1 e0 Hrece0|
   i e1 Hrece1 e0 Hrece0|
   e1 Hrece1 e0 Hrece0 e2 Hrece2|
   e1 Hrece1 e0 Hrece0|
   e Hrece|
   e Hrece|
   e1 Hrece1 e0 Hrece0|
   e1 Hrece1 e0 Hrece0]; intros x t t' envx envl Hnot_free Htyp;
 inversion Htyp.

(*const*)
apply type_of_const.

(*var*)
case (eq_A_dec x i).
intros x_is_i.
absurd (free_ident x (Var i)).
assumption.

rewrite x_is_i.
apply free_var.

intro x_isnt_i.
apply type_of_var.
apply apply_add_non_bis with (a := x) (b := t); assumption.

(*loc*)
apply type_of_loc; assumption.

(*fun*)
apply type_of_fun.
case (eq_A_dec x i).
intros x_is_i.
rewrite x_is_i in H4.
apply equiv_in_type_of with (envx' := add_table i t0 (add_table i t envx)).
apply equseless2; apply eqrefl.
assumption.

intro x_isnt_i.
apply Hrece with (x := x) (t := t).
intro Hfree.
absurd (free_ident x (Fun i e)).
assumption.

apply free_fun; assumption.

apply equiv_in_type_of with (envx' := add_table i t0 (add_table x t envx)).
apply eqswap. 
assumption. 
apply eqrefl.
assumption.


(*apply*)
apply type_of_app with (t := t0).
apply Hrece1 with (x := x) (t := t).
intro Hfree.
absurd (free_ident x (Apply e1 e0)).
assumption.

apply free_apply1; assumption.

assumption.

apply Hrece0 with (x := x) (t := t).
intro Hfree.
absurd (free_ident x (Apply e1 e0)).
assumption.

apply free_apply2; assumption.

assumption.

(*let*)
apply type_of_let with (t := t0).
apply Hrece1 with (x := x) (t := t).
intro Hfree.
absurd (free_ident x (Let_in i e1 e0)).
assumption.

apply free_let1; assumption.

assumption.

case (eq_A_dec x i).
intros x_is_i.
rewrite x_is_i in H6.
apply equiv_in_type_of with (envx' := add_table i t0 (add_table i t envx)).
apply equseless2; apply eqrefl.
assumption.

intro x_isnt_i.
apply Hrece0 with (x := x) (t := t).
intro Hfree.
absurd (free_ident x (Let_in i e1 e0)).
assumption.

apply free_let2; assumption.

apply equiv_in_type_of with (envx' := add_table i t0 (add_table x t envx)).
apply eqswap. 
assumption.
apply eqrefl.
assumption.

(*if*)
apply type_of_if with (t := t').
apply Hrece1 with (x := x) (t := t).
intro Hfree.
absurd (free_ident x (If e1 e0 e2)).
assumption.
apply free_if1; assumption.
assumption.

apply Hrece0 with (x := x) (t := t).
intro Hfree.
absurd (free_ident x (If e1 e0 e2)).
assumption.
apply free_if2; assumption.
assumption.

apply Hrece2 with (x := x) (t := t).
intro Hfree.
absurd (free_ident x (If e1 e0 e2)).
assumption.
apply free_if3; assumption.
assumption.

(*while*)
apply type_of_while with (t := t0).
apply Hrece1 with (x := x) (t := t).
intro Hfree.
absurd (free_ident x (While e1 e0)).
assumption.
apply free_while1; assumption.
assumption.

apply Hrece0 with (x := x) (t := t).
intro Hfree.
absurd (free_ident x (While e1 e0)).
assumption.
apply free_while2; assumption.
assumption.

(*ref*)
apply type_of_ref.
apply Hrece with (x := x) (t := t).
intro Hfree.
absurd (free_ident x (Ref e)).
assumption.
apply free_ref; assumption.
assumption.

(*deref*)
apply type_of_deref.
apply Hrece with (x := x) (t := t).
intro Hfree.
absurd (free_ident x (Deref e)).
assumption.
apply free_deref; assumption.
assumption.

(*sequ*)
apply type_of_sequ with (t' := t'0).
apply Hrece1 with (x := x) (t := t).
intro Hfree.
absurd (free_ident x (Sequ e1 e0)).
assumption.
apply free_sequ1; assumption.
assumption.

apply Hrece0 with (x := x) (t := t).
intro Hfree.
absurd (free_ident x (Sequ e1 e0)).
assumption.
apply free_sequ2; assumption.
assumption.

(*assign*)
apply type_of_assign with (t := t0).
apply Hrece1 with (x := x) (t := t).
intro Hfree.
absurd (free_ident x (Assign e1 e0)).
assumption.
apply free_assign1; assumption.
assumption.

apply Hrece0 with (x := x) (t := t).
intro Hfree.
absurd (free_ident x (Assign e1 e0)).
assumption.
apply free_assign2; assumption.
assumption.
Qed.