(********************************************************)
(*   Olivier Boite (under supervision C. Dubois) - 2001 *)
(*   Catherine Dubois - translation to V8 - sept. 2012  *)
(********************************************************)

Require Import Language_spec.
Require Import Table.
Require Import Typing_spec.
Require Import Small_step_spec.
Require Import Subject_reduction_proof.
Require Import Normal_form_proof.



Inductive red_star : configuration -> configuration -> Prop :=
  | red_0 : forall sn : configuration, red_star sn sn
  | red_n :
      forall sn s0 s' : configuration,
      red s0 s' -> red_star s' sn -> red_star s0 sn.

(*--------------------------------*) 
Lemma extend_trans :
 forall e1 e2 e3 : env_typ, extend e1 e2 -> extend e2 e3 -> extend e1 e3.

intros e1 e2 e3.
elim e1.
intros ex1 el1.
elim e2.
intros ex2 el2.
elim e3.
intros ex3 el3.
intros H12 H23.
inversion H12.
inversion H23.
rewrite H.
rewrite H1.
split.
trivial.

intros l t.
elim (H0 l t).
intros dome2e1 appe1e2.
elim (H2 l t).
intros dome3e2 appe2e3.
split.
intro.
apply dome2e1; apply dome3e2; assumption.

intro; apply appe1e2; apply appe2e3; assumption.
Qed.

 


(*--------------------------------*) 
Lemma less_typable_trans :
 forall s0 s sn : configuration,
 less_typable s0 s -> less_typable s sn -> less_typable s0 sn.
intros.
unfold less_typable in |- *.
intros env t Htyp.
unfold less_typable in H.
cut (exists env' : env_typ, extend env' env /\ type_of_config env' s t).
intro Hex_env'_s.
elim Hex_env'_s.
intro env's.
intros Hconj.
clear H.
unfold less_typable in H0.
cut
 (exists env'sn : env_typ, extend env'sn env's /\ type_of_config env'sn sn t).
intro Hex_env'sn.
elim Hex_env'sn.
intro env'sn.
clear H0.
intro Hconj2.
split with env'sn.
split.
inversion Hconj2.
inversion Hconj.
apply extend_trans with (e2 := env's); assumption.

inversion Hconj2; assumption.

apply H0.
inversion Hconj; assumption.

apply H.
assumption.
Qed.




(*--------------------------------*) 
Lemma subject_reduction_star :
 forall s0 sn : configuration, red_star s0 sn -> less_typable s0 sn.
intros.
induction H as [sn| sn s0 s' H H0 HrecH].
unfold less_typable in |- *.
intros env t Htyp.
split with env.
split.
apply extend_id.

assumption.

apply less_typable_trans with (s := s').
apply Subject_reduction; assumption.

assumption.
Qed.


(*--------------------------------*) 
Definition config_is_value (s : configuration) :=
  match s with
  | (a, m) => is_value a
  end.


(*--------------------------------*) 
Theorem type_safe :
 forall (a : expr) (t : type) (s' : configuration) 
   (fm : location -> value) (fx : identifier -> type) 
   (fl : location -> type),
 red_star (a, empty_table fm) s' ->
 type_of_config (empty_table fx, empty_table fl) (a, empty_table fm) t ->
 ~ is_reducible s' ->
 config_is_value s' /\ (exists env : env_typ, type_of_config env s' t).



intros a t s' fm fx fl Hred Htyp. 
set (env := (empty_table fx, empty_table fl)) in Htyp.
cut (exists env' : env_typ, extend env' env /\ type_of_config env' s' t).
induction Hred as [sn| sn s0 s' H Hred HrecHred].


(*red_0*)
elim sn.
intros an mn HEX Hnf.
split.
simpl in |- *.
inversion_clear HEX.
inversion_clear H.
generalize H0.
generalize H1.
clear H0.
clear H1.
elim x.
clear x.
intros envx' envl' Htyp0 Hext.
unfold env in Hext.
apply (nf_typed_are_value fx envl' an t mn).
inversion Hext.
rewrite H in Htyp0.
assumption.

assumption.

inversion_clear HEX.
inversion_clear H.
exists x.
assumption.

intros.
inversion_clear H0.
inversion_clear H2.
split.
destruct sn; simpl.
destruct x.
unfold env in H0;inversion H0.
eapply nf_typed_are_value with (m:=m) (t:=t) (f:=fx) (envl:=e1).
rewrite H2 in H3;assumption.
assumption.
exists x; assumption.

cut (less_typable (a, empty_table fm) s').
intro Hlesstypable.

apply Hlesstypable.
assumption.

apply subject_reduction_star.
assumption.
Qed.
